/*		 ICP-PCI Kernel module free public release
 *
 *		       Copyright (C) 1996-99 GOM mbH.
 *	         Rebenring 33, 38106 Braunschweig, GERMANY
 *		            All rights reserved
 *
 * $Id: icpciAux.c,v 5.1.2.7 1999/09/22 11:47:37 matthias Exp $
 *
 */

/*
 * Includes:
 */

#include "icpciOS.h"	/* os specific stuff */
#include "icpciExt.h"	/* offical driver interface */
#include "icpciInt.h"	/* driver internals */
#include "icpciReg.h"	/* register definitions IC-PCI */
#include "amvsReg.h"	/* register definitions AM-VS */
#include "amdigReg.h"	/* register definitions AM-DIG */
#include "ampcvReg.h"	/* register definitions PCVision */

/*
 * Defines:
 */

/*
 * Globals:
 */

/* human readable error messages */
const char *icp_errmsg[] =
{ "no error",
  "mask error",	
  "range error",
  "operation not supported",
  "null pointer access",
  "operation already in use"
};

extern struct 
icp_global_t    icp;	/* main driver structure */

struct pv_reg_t
*pv_reg[ICP_BOARDS_MAX];/* pointer to PCVision registers */

extern struct vs_cam_t vs_def_cam;	/* \               */
extern struct dg_cam_t dg_def_cam;	/* - see icpciAM.c */
extern struct pv_cam_t pv_def_cam;	/* /               */

/*
 * Prototypes:
 */

/*
 * special AMCC- REV. 0 handling
 */
int
icp_check_amcc (int board, void *address, u_char operation) {

  int ret_val;

  ret_val = FALSE;
  if(icp.cap[board]->latch_status == ICP_AMCC_ODD && ((u_long)address & 2)) {
    /* odd read/write followed by even read/write */
    ret_val = TRUE;
  } else {
    /* even read followed by odd write */
    if(icp.cap[board]->latch_status == ICP_AMCC_EVEN &&
       icp.cap[board]->last_operation == ICP_AMCC_READ &&
       ((u_long)address & 2) && operation == ICP_AMCC_WRITE) {
      ret_val = TRUE;
    }
  }
  icp.cap[board]->latch_status =
    ((u_long)address & 2) ? ICP_AMCC_ODD : ICP_AMCC_EVEN;
  icp.cap[board]->last_operation = operation;

  return ret_val;
}

/*
 * icp_error
 */

void
icp_error (u_char board, const char* function)
{
  
  icp_printf ("icpci[%d]: %s: %s\n", board, function,
	      icp_errmsg[-icp.errno[board]]);
  icp.errno[board] = 0;
}

/*
 *  icp_decode_devno
 */

int
icp_decode_devno (struct icp_devno_t *devno)
{
  
  /* decode */
  devno->board = (devno->minor & ICP_BRD_MASK) >> ICP_BRD_SHIFT;
  devno->module = (devno->minor & ICP_MOD_MASK) >> ICP_MOD_SHIFT;
  devno->camera = (devno->minor & ICP_CAM_MASK) >> ICP_CAM_SHIFT;
  
  /* fixup the module type for the PCVision
     means remap from 3 to 16 */
  
  if (devno->major == devno->major1 && devno->module == 3 )
    devno->module = ICP_AMPV;
  
  /* debugging */ 
  ICP_PRINTF("icpci[%c]: board: %d, module: %d, camera: %d.\n",
	     'A', devno->board, devno->module, devno->camera);
  
  /* device OK? */

  /* board request valid? */
  if (devno->board >= icp.boards)
    return -ENXIO;

  /* module request valid? */
  if (devno->module != ICP_MOD_IDENT)
    {
      if (devno->module != ICP_AMNO)
	/* if you don't want to access the io-ports */
	if (devno->module != icp.cap[devno->board]->module)
	  return -ENXIO;
    }
  else
    {
      /* on self identifying devices only special camera device valid */
      if (devno->camera != ICP_CAM_IOCTL)
	return -ENXIO;
    }

  /* camera request valid? */
  if (devno->camera != ICP_CAM_IOCTL)
    {
      if (devno->module == ICP_AMNO)
	{
	  /* you can access the io-ports only on camera 0 */
	  if (devno->camera != 0)
	    return -ENXIO;
	}
      else
	{
	  if (devno->camera >= icp.cap[devno->board]->cameras)
	    return -ENXIO;
	}
    }

  return OK;
}

/*
 * New Interrupt Handler Interface
 */

int
icp_enable_irq (int board, u_long irqs)
{
  int shift;

  /* disable irq's */  
  icp_irqoff (board);

  /* clear all requests */
  xcr_clear_irqs (board);

  /* enable irq's in this function */
  shift = 0;
  while (irqs)
    {
      if (shift < 16)
	switch (shift)
	  {
	  case 0:
	    if (icp.cap[board]->module == ICP_AMPV)
	      {
		if (irqs & 1) {
		  if (pv_set_sotien (board, PV_ENABLE) < 0) {
		    icp_error (board, "pv_set_sotien");
		    return -EIO;
		  }
		} else {
		  if (pv_set_sotien (board, PV_DISABLE) < 0) {
		    icp_error (board, "pv_set_sotien");
		    return -EIO;
		  }
		}
	      }
	    break;
	  case 1:
	    if (icp.cap[board]->module == ICP_AMPV)
	      {
		if (irqs & 1) {
		  if (pv_set_eotien (board, PV_ENABLE) < 0) {
		    icp_error (board, "pv_set_eotien");
		    return -EIO;
		  }
		} else {
		  if (pv_set_eotien (board, PV_DISABLE) < 0) {
		    icp_error (board, "pv_set_eotien");
		    return -EIO;
		  }
		}
	      }
	    break;
	  case 2:
	    if (icp.cap[board]->module == ICP_AMPV)
	      {
		if (irqs & 1) {
		  if (pv_set_ein0ien (board, PV_ENABLE) < 0) {
		    icp_error (board, "pv_set_ein0ien");
		    return -EIO;
		  }
		} else {
		  if (pv_set_ein0ien (board, PV_DISABLE) < 0) {
		    icp_error (board, "pv_set_ein0ien");
		    return -EIO;
		  }
		}
	      }
	    break;
	  case 3:
	    if (icp.cap[board]->module == ICP_AMPV)
	      {
		if (irqs & 1) {
		  if (pv_set_ein1ien (board, PV_ENABLE) < 0) {
		    icp_error (board, "pv_set_ein1ien");
		    return -EIO;
		  }
		} else {
		  if (pv_set_ein1ien (board, PV_DISABLE) < 0) {
		    icp_error (board, "pv_set_ein1ien");
		    return -EIO;
		  }
		}
	      }
	    break;
	  case 4:
	    if (icp.pci[board]->revision > ICP_REV_A)
	      {
		if (irqs & 1) {
		  if (cr_set_inportien (board, CR_ENABLE) < 0) {
		    icp_error (board, "cr_set_inportien");
		    return -EIO;
		  }
		} else {
		  /* the io-port interrupt is not disabled by default! */
		}
	      }
	    break;
	  case 7:
	    if (irqs & 1) {
	      if (cr_set_vbien (board, CR_ENABLE) < 0) {
		icp_error (board, "cr_set_vbien");
		return -EIO;
	      }
	    } else {
	      if (cr_set_vbien (board, CR_DISABLE) < 0) {
		icp_error (board, "cr_set_vbien");
		return -EIO;
	      }
	    }
	    break;
	  case 8:
	    if (irqs & 1) {
	      if (cr_set_ovbien (board, CR_ENABLE) < 0) {
		icp_error (board, "cr_set_ovbien");
		return -EIO;
	      }
	    } else {
	      if (cr_set_ovbien (board, CR_DISABLE) < 0) {
		icp_error (board, "cr_set_ovbien");
		return -EIO;
	      }
	    }
	    break;
	  case 9:
	    if (irqs & 1) {
	      if (cr_set_evbien (board, CR_ENABLE) < 0) {
		icp_error (board, "cr_set_evbien");
		return -EIO;
	      }
	    } else {
	      if (cr_set_evbien (board, CR_DISABLE) < 0) {
		icp_error (board, "cr_set_evbien");
		return -EIO;
	      }
	    }
	    break;
	  case 10:
	    if (irqs & 1) {
	      if (cr_set_amien (board, CR_ENABLE) < 0) {
		icp_error (board, "cr_set_amien");
		return -EIO;
	      }
	    } else {
	      if (cr_set_amien (board, CR_DISABLE) < 0) {
		icp_error (board, "cr_set_amien");
		return -EIO;
	      }
	    }
	    break;
	  case 11:
	    if (irqs & 1) {
	      if (cr_set_trigien (board, CR_ENABLE) < 0) {
		icp_error (board, "cr_set_trigien");
		return -EIO;
	      }
	    } else {
	      if (cr_set_trigien (board, CR_DISABLE) < 0) {
		icp_error (board, "cr_set_trigien");
		return -EIO;
	      }
	    }
	    break;
	  case 12:
	    if (irqs & 1) {
	      if (cr_set_adrien (board, CR_ENABLE) < 0) {
		icp_error (board, "cr_set_adrien");
		return -EIO;
	      }
	    } else {
	      if (cr_set_adrien (board, CR_DISABLE) < 0) {
		icp_error (board, "cr_set_adrien");
		return -EIO;
	      }
	    }
	    break;
	  case 13:
	    if (irqs & 1) {
	      if (cr_set_aqien (board, CR_ENABLE) < 0) {
		icp_error (board, "cr_set_aqien");
		return -EIO;
	      }
	    } else {
	      if (cr_set_aqien (board, CR_DISABLE) < 0) {
		icp_error (board, "cr_set_aqien");
		return -EIO;
	      }
	    }
	    break;
	  case 14:
	    if (irqs & 1) {
	      switch (cr_get_fldsel (board)) {
	      case CR_EVEN_FIELD:
		if (cr_set_vbien (board, CR_DISABLE) < 0) {
		  icp_error (board, "cr_set_vbien");
		  return -EIO;
		}
		if (cr_set_evbien (board, CR_ENABLE) < 0) {
		  icp_error (board, "cr_set_evbien");
		  return -EIO;
		}
		if (cr_set_ovbien (board, CR_DISABLE) < 0) {
		  icp_error (board, "cr_set_ovbien");
		  return -EIO;
		}
		break;
	      case CR_ODD_FIELD:
		if (cr_set_vbien (board, CR_DISABLE) < 0) {
		  icp_error (board, "cr_set_vbien");
		  return -EIO;
		}
		if (cr_set_evbien (board, CR_DISABLE) < 0) {
		  icp_error (board, "cr_set_evbien");
		  return -EIO;
		}
		if (cr_set_ovbien (board, CR_ENABLE) < 0) {
		  icp_error (board, "cr_set_ovbien");
		  return -EIO;
		}
		break;
	      case CR_NEXT_FIELD:
		if (cr_set_vbien (board, CR_ENABLE) < 0) {
		  icp_error (board, "cr_set_vbien");
		  return -EIO;
		}
		if (cr_set_evbien (board, CR_DISABLE) < 0) {
		  icp_error (board, "cr_set_evbien");
		  return -EIO;
		}
		if (cr_set_ovbien (board, CR_DISABLE) < 0) {
		  icp_error (board, "cr_set_ovbien");
		  return -EIO;
		}
		break;
	      default:
		return -EINVAL;
	      }
	    } else {
	      if (cr_set_vbien (board, CR_DISABLE) < 0) {
		icp_error (board, "cr_set_vbien");
		return -EIO;
	      }
	      if (cr_set_ovbien (board, CR_DISABLE) < 0) {
		icp_error (board, "cr_set_ovbien");
		return -EIO;
	      }
	      if (cr_set_evbien (board, CR_DISABLE) < 0) {
		icp_error (board, "cr_set_evbien");
		return -EIO;
	      }
	    }
	    break;
	  default:
	    if (irqs & 1) {
	      return -EINVAL;
	    }
	  }
      else
	if ((irqs & 1) == ICR_ENABLE)
	  switch (shift) {
	  case 17:
	    if (icr_set_inten (board, ICR_ENABLE) < 0) {
	      icp_error (board, "cr_set_inten");
	      return -EIO;
	    }
	    break;
	  case 18:
	    if (icp.bma_on) {
	      if (icr_set_binten (board, ICR_ENABLE) < 0) {
		icp_error (board, "icr_set_binten");
		return -EIO;
	      }
	    }
	    break;
	  default:
	    return -EINVAL;
	  }
      irqs >>= 1;
      shift++;
    }

  /* clear all requests */
  xcr_clear_irqs (board);

  /* enable irq's */
  icp_irqon (board);

  return (OK);
}

int
icp_disable_irq (int board, u_long irqs)
{
  int shift;

  /* disable irq's */
  icp_irqoff (board);

  /* clear all requests */
  xcr_clear_irqs (board);

  /* disable irq's in this function */
  shift = 0;
  while (irqs) {
    if (shift < 16)
      {
	if ((irqs & 1) == CR_ENABLE)
	  switch (shift)
	    {
	    case 0:
	      if (icp.cap[board]->module == ICP_AMPV) {
		if (pv_set_sotien (board, CR_DISABLE) < 0) {
		  icp_error (board, "pv_set_sotien");
		  return -EIO;
		}
	    }
	      break;
	    case 1:
	      if (icp.cap[board]->module == ICP_AMPV) {
		if (pv_set_eotien (board, CR_DISABLE) < 0) {
		  icp_error (board, "pv_set_eotien");
		  return -EIO;
		}
	      }
	      break;
	    case 2:
	      if (icp.cap[board]->module == ICP_AMPV) {
		if (pv_set_ein0ien (board, CR_DISABLE) < 0) {
		  icp_error (board, "pv_set_ein0ien");
		  return -EIO;
		}
	      }
	      break;
	    case 3:
	      if (icp.cap[board]->module == ICP_AMPV) {
		if (pv_set_ein1ien (board, CR_DISABLE) < 0) {
		icp_error (board, "pv_set_ein1ien");
		return -EIO;
		}
	      }
	      break;
	    case 4:
	      if (icp.pci[board]->revision > ICP_REV_A)
		if (cr_set_inportien (board, CR_DISABLE) < 0) {
		  icp_error (board, "cr_set_inportien");
		  return -EIO;
		}
	      break;
	    case 7:
	      if (cr_set_vbien (board, CR_DISABLE) < 0) {
		icp_error (board, "cr_set_vbien");
		return -EIO;
	      }
	      break;
	    case 8:
	      if (cr_set_ovbien (board, CR_DISABLE) < 0) {
		icp_error (board, "cr_set_ovbien");
		return -EIO;
	      }
	      break;
	    case 9:
	      if (cr_set_evbien (board, CR_DISABLE) < 0) {
		icp_error (board, "cr_set_evbien");
		return -EIO;
	      }
	      break;
	    case 10:
	      if (cr_set_amien (board, CR_DISABLE) < 0) {
		icp_error (board, "cr_set_amien");
		return -EIO;
	      }
	      break;
	    case 11:
	      if (cr_set_trigien (board, CR_DISABLE) < 0) {
		icp_error (board, "cr_set_trigien");
		return -EIO;
	      }
	      break;
	    case 12:
	      if (cr_set_adrien (board, CR_DISABLE) < 0) {
		icp_error (board, "cr_set_adrien");
		return -EIO;
	      }
	      break;
	    case 13:
	      if (cr_set_aqien (board, CR_DISABLE) < 0) {
		icp_error (board, "cr_set_aqien");
		return -EIO;
	      }
	      break;
	    case 14:
	      if (cr_set_vbien (board, CR_DISABLE) < 0) {
		icp_error (board, "cr_set_vbien");
		return -EIO;
	      }
	      if (cr_set_ovbien (board, CR_DISABLE) < 0) {
		icp_error (board, "cr_set_ovbien");
		return -EIO;
	      }
	      if (cr_set_evbien (board, CR_DISABLE) < 0) {
		icp_error (board, "cr_set_evbien");
		return -EIO;
	      }
	      break;
	    default:
	      return -EINVAL;
	    }
      }
    else
      {
	if ((irqs & 1) == ICR_ENABLE)
	  switch (shift) {
	  case 17:
	    if (icr_set_inten (board, ICR_DISABLE) < 0) {
	      icp_error (board, "cr_set_inten");
	      return -EIO;
	    }
	    break;
	  case 18:
	    if (icr_set_binten (board, ICR_DISABLE) < 0) {
	      icp_error (board, "icr_set_binten");
	      return -EIO;
	    }
	    break;
	  default:
	    return -EINVAL;
	  }
      }
    irqs >>= 1;
    shift++;
  }

  /* enable irq's */
  icp_irqon (board);

  return (OK);
}

/*
 * icp_icr_init()
 */

int
icp_icr_init (int board) {

  ICP_PRINTF("icpci[%c]: icp_icr_init.\n", 'F');

  icp.errno[board] = EICP_ERR_NO;

  /* initialize registers */

  icr_set_mbox1 (board, ICR_MBOX_INIT);
  icr_set_mbox2 (board, ICR_MBOX_INIT);
  icr_set_mbox3 (board, ICR_MBOX_INIT);
  icr_set_mbox4 (board, ICR_MBOX_INIT);
  icr_set_bmdst (board, ICR_BMDST_INIT);
  icr_set_bmxc (board, ICR_BMXC_INIT);
  icr_init_intctl (board);
  icr_init_bmctl (board);

  /* check for errors */
  
  if (icp.errno[board] < 0)
    {
    icp_error (board, "icp_icr_init");
    return -EIO;
    }

  return OK;
}

/*
 * icp_cr_init()
 */

int
icp_cr_init (int board) {

  ICP_PRINTF("icpci[%c]: icp_cr_init.\n", 'F');

  icp.errno[board] = EICP_ERR_NO;

  /* initialize registers */

  cr_init_con (board);
  cr_init_acqcon (board);
  cr_set_pitch (board, CR_PITCH_INIT);
  cr_set_aoix (board, CR_AOIX_INIT);
  cr_set_bmsf (board, CR_BMSF_INIT);
  cr_set_bmss (board, CR_BMSS_INIT);
  cr_set_aqstart (board, CR_AQSTART_INIT);
  cr_set_intadr (board, CR_INTADR_INIT);
  cr_init_intcon (board);
  cr_init_mdata (board);

  /* extended init on REV-B boards (including PCVision) */

  if (icp.pci[board]->revision > ICP_REV_A) {
    cr_init_wzoom (board);
    cr_init_portcon (board);
    cr_init_outport (board);
    cr_init_inport (board);
  }

  /* check for errors */

  if (icp.errno[board] < 0)
    {
      icp_error (board, "icp_cr_init");
      return -EIO;
    }

  /* set right memory size */

  if (icp.map[board]->big_mem)
    {
      if (cr_set_memcfg (board, CR_MEM_4) < 0)
	{
	  icp_error (board, "cr_set_mem");
	  return -EIO;
	}
    } 
  else
    {
      if (cr_set_memcfg (board, CR_MEM_2) < 0)
	{
	icp_error (board, "cr_set_mem");
	return -EIO;
	}
    }
  
  return OK;
}

/*
 * icp_pci_init
 */

int
icp_pci_init (int *boards)
{
  int board, error;
  
  ICP_PRINTF("icpci[%c]: icp_pci_init.\n", 'F');
  
  if ((error = icp_os_pci_init (boards)))
    return error;
  
  if (icp.bma_on)
    {
      ICP_PRINTF("icpci[%c]: Upper Memory    at: %#lx.\n",
		 'I', icp.p_bma_base);
      ICP_PRINTF("icpci[%c]: Remapping (UM)  at: 0x%p.\n",
		 'I', icp.v_bma_base);
      ICP_PRINTF("icpci[%c]:                   : [%ld page(s)].\n",
		 'I', icp.bma_size / PAGE_SIZE); 
    }

  for (board = 0; board < *boards; board++)
    {
      /* debugging output of pci config space */
      ICP_PRINTF("icpci[%d]: Interrupt : %d.\n", board,
		 icp.pci[board]->irq_line);
      ICP_PRINTF("icpci[%d]: Revision  : %d.\n", board,
		 icp.pci[board]->revision & ~ICP_MEMSZ_MASK); 
      ICP_PRINTF("icpci[%d]: ICR Register Set at: %#lx.\n",
		 board, icp.pci[board]->address0);
      ICP_PRINTF("icpci[%d]: CR Register Set  at: %#lx.\n",
		 board, icp.pci[board]->address1);
      ICP_PRINTF("icpci[%d]: AM Register Set  at: %#lx.\n",
		 board, icp.pci[board]->address2);
      ICP_PRINTF("icpci[%d]: On Board Memory  at: %#lx.\n",
		 board, icp.pci[board]->address3);
    }

  return OK;
}

/*
 * icp_map_init
 */

int
icp_map_init (int board) {

  int error;

  u_long align_div[NUM_REGS];
  ptrdiff_t align_mod[NUM_REGS];

  ICP_PRINTF("icpci[%c]: icp_map_init.\n", 'F');
  
  /* allocate structure */
  icp.map[board] =
    (struct icp_map_t*)icp_malloc (sizeof (struct icp_map_t), UNPAGED);
  if (icp.map[board] == NULL)
    return -ENOMEM;
  memset (icp.map[board], 0x00, sizeof(struct icp_map_t));
  
  /* memory size is decoded in revision id */
  icp.map[board]->big_mem = (icp.pci[board]->revision & ICP_MEMSZ_MASK) >>
    ICP_MEMSZ_SHIFT;
  icp.pci[board]->revision &= ~ICP_MEMSZ_MASK;
  
  /* PCVision has always 2Mb of on board memory */
  if (icp.pci[board]->revision > PV_REV_MIN &&
      icp.map[board]->big_mem)
    {
      icp_printf("icpci[%d]: PCVision with 4Mb memory detected?\n", board);
      return -EIO;
    }
  
  /* are cr/icr registers mmaped or io-ports */
  if ((icp.pci[board]->address0 & ICP_PCI_BASE_ADDR_SPACE) ==
      ICP_PCI_BASE_ADDR_SPACE_IO)
    {
      icp.pci[board]->address0 &= ICP_PCI_BASE_ADDR_IO_MASK;
      icp.map[board]->regs_io = TRUE;
    } 
  else
    {
      icp.pci[board]->address0 &= ICP_PCI_BASE_ADDR_MEM_MASK;
      icp.map[board]->regs_io = FALSE;
    }
  
  /* cr and icr have to be accessed in the same way (mmaped or io) */
  if (icp.map[board]->regs_io)
    {
      if ((icp.pci[board]->address1 & ICP_PCI_BASE_ADDR_SPACE) !=
	  ICP_PCI_BASE_ADDR_SPACE_IO)
	{
	  icp_printf("icpci[%d]: control regs not accessible via i/o ops.\n",
		     board);
	  return -EIO;
	}
      else
	{
	  icp.pci[board]->address1 &= ICP_PCI_BASE_ADDR_IO_MASK;
	  ICP_PRINTF("icpci[%d]: ICR/CR Registers are i/o ports.\n", board);
	}
    }
  else
    {
      if ((icp.pci[board]->address1 & ICP_PCI_BASE_ADDR_SPACE) !=
	  ICP_PCI_BASE_ADDR_SPACE_MEMORY)
	{
	  icp_printf("icpci[%d]: control registers not mmaped \n", board);
	  return -EIO;
	}
      else
	{
	  icp.pci[board]->address1 &= ICP_PCI_BASE_ADDR_MEM_MASK;
	  ICP_PRINTF("icpci[%d]: ICR/CR Registers are mmaped.\n", board);
	}
    }

  /* acquisition control regs are always mmaped */
  if ((icp.pci[board]->address2 & ICP_PCI_BASE_ADDR_SPACE) !=
      ICP_PCI_BASE_ADDR_SPACE_MEMORY)
    {
      icp_printf("icpci[%d]: acquisition module registers not mmaped.\n",
		 board);
      return -EIO;
    }
  else
    icp.pci[board]->address2 &= ICP_PCI_BASE_ADDR_MEM_MASK;
  
  /* frame-grabber memory too */
  if ((icp.pci[board]->address3 & ICP_PCI_BASE_ADDR_SPACE) !=
      ICP_PCI_BASE_ADDR_SPACE_MEMORY)
    {
      icp_printf("icpci[%d]: frame grabber on board  memory not available.\n",
		 board);
      return -EIO;
    }
  else
    icp.pci[board]->address3 &= ICP_PCI_BASE_ADDR_MEM_MASK;
  
  /* align icr/cr register sets to page boundaries (needed for ami bios'es) */
  if (!icp.map[board]->regs_io)
    {
      if ((align_mod[ICR_REGS] = icp.pci[board]->address0 % PAGE_SIZE))
	ICP_PRINTF("icpci[%c]: interface control regs not aligned.\n", 'I');
      align_div[ICR_REGS] = ROUND_DOWNTO_PAGE(icp.pci[board]->address0);
      if (align_div[ICR_REGS] !=
	  ROUND_DOWNTO_PAGE(icp.pci[board]->address0 +
			    ICR_REG_SIZE * sizeof(u_long) - 1))
	ICP_PRINTF("icpci[%c]: icr reg-space crosses page boundary.\n", 'I');
      
      if ((align_mod[CR_REGS] = icp.pci[board]->address1 % PAGE_SIZE))
	ICP_PRINTF("icpci[%c]: control regs not aligned.\n", 'I');
      align_div[CR_REGS] = ROUND_DOWNTO_PAGE(icp.pci[board]->address1);
      if (align_div[CR_REGS] !=
	  ROUND_DOWNTO_PAGE(icp.pci[board]->address1 +
			    CR_REG_SIZE * sizeof(u_short) - 1))
	ICP_PRINTF("icpci[%c]: cr reg-space crosses page boundary.\n", 'I');
    }
  
  /* module register set and on board memory are always aligned */
  if ((icp.pci[board]->address2 % PAGE_SIZE))
    {
      icp_printf("icpci[%c]: acquisition modul regs not aligned.\n", 'E');
      return -EFAULT;
    }
  
  if ((icp.pci[board]->address3 % PAGE_SIZE))
    {
      icp_printf("icpci[%d]: main board memory not aligned.\n", 'E');
      return -EFAULT;
    }

  
  if ((error = icp_os_map_init (board, align_div, align_mod)))
    return error;
  
  /* debugging output of pci config space */
  if (!icp.map[board]->regs_io)
    {
      ICP_PRINTF("icpci[%d]: Remapping (ICR) at: 0x%p.\n",
		 board, icp.map[board]->v_icr_base);
      ICP_PRINTF("icpci[%d]:             offset: %#x.\n",
		 board, align_mod[ICR_REGS]);
      ICP_PRINTF("icpci[%d]:                   : [%ld page(s)].\n",
		 board, ROUND_UPTO_PAGE(align_mod[ICR_REGS] +
					ICR_REG_SIZE * sizeof(u_long)) /
		 PAGE_SIZE);
      ICP_PRINTF("icpci[%d]: Remapping (CR)  at: 0x%p.\n",
		 board, icp.map[board]->v_cr_base);
      ICP_PRINTF("icpci[%d]:             offset: %#x.\n",
		 board, align_mod[CR_REGS]);
      ICP_PRINTF("icpci[%d]:                   : [%ld page(s)].\n",
		 board, ROUND_UPTO_PAGE(align_mod[CR_REGS] +
					CR_REG_SIZE * sizeof(u_short)) /
		 PAGE_SIZE);
    }
  ICP_PRINTF("icpci[%d]: Remapping (AM)  at: 0x%p.\n",
	     board, icp.map[board]->v_mod_base);
  ICP_PRINTF("icpci[%d]:                   : [%ld page(s)].\n",
	     board, MOD_MAP_SIZE / PAGE_SIZE);
  ICP_PRINTF("icpci[%d]: Remapping (BM)  at: 0x%p.\n",
	     board, icp.map[board]->v_mem_base);
  ICP_PRINTF("icpci[%d]:                   : [%ld page(s)].\n",
	     board, icp.map[board]->big_mem ?
	     (MEM_MAP_SIZE << 1) / PAGE_SIZE :
	     MEM_MAP_SIZE / PAGE_SIZE);
  
  /* fixup the possibly unaligned base addresses */
  if (icp.map[board]->regs_io)
    {
      icp.map[board]->v_icr_base = NULL;
      icp.map[board]->v_cr_base = NULL;
      icp.map[board]->io_icr_base = icp.pci[board]->address0;
      icp.map[board]->io_cr_base = icp.pci[board]->address1;
    }
  else
    {
      icp.map[board]->v_icr_base += align_mod[ICR_REGS] / sizeof(u_long);
      icp.map[board]->v_cr_base += align_mod[CR_REGS] / sizeof(u_short);
      icp.map[board]->io_icr_base = 0UL;
      icp.map[board]->io_cr_base = 0UL;
    }

  return OK;
}

/*
 *  icp_get_modid
 */

static u_char
icp_get_modid (u_char board) {
  
  u_short tmp_val;
  
  tmp_val = icp_readw (icp.map[board]->v_mod_base + ICP_MODID_OFFST);
  
  return ((tmp_val & ICP_TYPE_MASK) >> ICP_TYPE_SHIFT);
}

/*
 * icp_cap_init
 */

int
icp_cap_init (int board)
{
  int error;

  ICP_PRINTF("icpci[%c]: icp_cap_init.\n", 'F');

  /* allocate structure */
  icp.cap[board] =
    (struct icp_cap_t*)icp_malloc (sizeof (struct icp_cap_t), UNPAGED);
  if (icp.cap[board] == NULL)
    return -ENOMEM;
  memset (icp.cap[board], 0x00, sizeof(struct icp_cap_t));

  /* test for various hardware bugs */
  if (icp.pci[board]->revision == 0)
    /* AMCC Rev0 Bug */
    {
      icp_printf ("icpci[%d]: AMCC PCI Matchmaker Rev. 0 detected!\n",
		  board);
      icp_printf ("           Workaround enabled!\n");
      icp.cap[board]->amcc0_bug = TRUE;
    }
  else
    /* AMCC Rev1 Bug (S5933Q vs. S5933QB)*/
    if ((icp.pci[board]->status & ICP_LEVEL_MASK) >> ICP_LEVEL_SHIFT)
      {
	icp_printf ("icpci[%d]: AMCC PCI Matchmaker Rev. 1Q detected!\n",
		    board);
	icp_printf ("           Workaround enabled!\n");
	icp.cap[board]->amcc1_bug = TRUE;
	if ((error = icp_os_cap_init (board)))
	  return error;
      }

  /* initialize the base register sets */
  if ((error = icp_icr_init (board)))
    {
      icp_printf ("icpci[%c]: Init. of Interface Control Regs failed.\n",
		  board);
      return error;
    }
  if ((error = icp_cr_init (board)))
    {
      icp_printf ("icpci[%c]: Init. of Control Regs failed.\n", board);  
      return error;
    }
  
  /* autodetect acquisition module and initialize it */
  if (!cr_get_amp(board))
    {
      icp.cap[board]->module = icp_get_modid (board);
      switch (icp.cap[board]->module)
	{
	case ICP_AMFA:
	  icp_printf ("icpci[%d]: AM-FA Acquisition Module found.\n", board);
	  icp_printf ("           Not supported yet!.\n");
	  return -ENXIO;
	  break;
	case ICP_AMC1:
	  icp_printf ("icpci[%d]: AM-C1 Acquisition Module found.\n", board);
	  icp_printf ("           Not supported yet!.\n");
	  return -ENXIO;
	  break;
	case ICP_AMMTD:
	  icp_printf ("icpci[%d]: AM-MTD Acquisition Module found.\n", board);
	  icp_printf ("           Not supported yet!.\n");
	  return -ENXIO;
	  break;
	case ICP_AMCMP:
	  icp_printf ("icpci[%d]: AM-STD/COMP Acquisition Module found.\n",
		      board);
	  icp_printf ("           Not supported yet!.\n");
	  return -ENXIO;
	  break;
	case ICP_AMRGB:
	  icp_printf ("icpci[%d]: AM-STD/RGB Acquisition Module found.\n",
		      board);
	  icp_printf ("           Not supported yet!.\n");
	  return -ENXIO;
	  break;
	case ICP_AMVS:
	  icp_printf ("icpci[%d]: AM-VS Acquisition Module found.\n", board);
	  if ((error = icp_amvs_startup (board)))
	    {
	      icp_printf ("icpci[%d]: Startup of AM-VS Registers failed.\n",
			  board);
	      return error;
	    }
	  icp.cap[board]->cameras = VS_CAMERAS_MAX;
	  break;
	case ICP_AMDG:
	  icp_printf ("icpci[%d]: AM-DIG Acquisition Module found.\n", board);
	  if ((error = icp_amdg_startup (board)))
	    {
	      icp_printf ("icpci[%d]: Startup of AM-DIG Registers failed.\n",
			  board);
	      return error;
	    }
	  icp.cap[board]->cameras = DG_CAMERAS_MAX;
	  break;
	case ICP_AMPV:
	  icp_printf ("icpci[%d]: PCVision Acquisition Module found.\n",
		      board);
	  pv_reg[board] = (struct pv_reg_t*)icp.map[board]->v_mod_base;
	  if ((error = icp_ampv_startup (board)))
	    {
	      icp_printf ("icpci[%d]: Startup of PCVision Registers failed.\n",
			  board);
	      return error;
	    }
	  icp.cap[board]->cameras = PV_CAMERAS_MAX;
	  break;
	default:
	  icp_printf ("icpci[%d]: Unknown Module Type (%d) found.\n", board,
		      icp.cap[board]->module);
	  return -ENXIO;
	}
    }
  else
    {
      icp_printf ("icpci[%d]: No Acquisition Module installed.\n", board);
      return -ENXIO;
    }
  
  /* no camera open yet */
  icp.cap[board]->cam_open = FALSE;
  
  return OK;
}

/*
 * icp_dev_init
 */

int
icp_dev_init (int board)
{
  int	error, camera;
  
  ICP_PRINTF("icpci[%c]: icp_dev_init.\n", 'F');
  
  /* set the default structure for every device */
  for (camera = 0; camera < icp.cap[board]->cameras; camera++) {
    
    /* allocate structure */
    icp.dev[board][camera] =
      (union icp_dev_t*)icp_malloc (sizeof (union icp_dev_t), UNPAGED);
    if (icp.dev[board][camera] == NULL)
      return -EFAULT;
    memset (icp.dev[board][camera], 0x00, sizeof(union icp_dev_t));
    
    switch (icp.cap[board]->module) {
    case ICP_AMVS:
      memcpy (&icp.dev[board][camera]->vs_cam,
	      &vs_def_cam, sizeof(struct vs_cam_t));
      if (camera == 0)
	if ((error = icp_amvs_init (board, 0)))
	  {
	    icp_printf ("icpci[%d]: Init. of AM-VS Registers failed.\n",
			board);
	    return error;
	  }
      break;
    case ICP_AMDG:
      memcpy (&icp.dev[board][camera]->dg_cam,
	      &dg_def_cam, sizeof(struct dg_cam_t));
      if ((error = icp_amdg_init (board, 0)))
	{
	  icp_printf ("icpci[%d]: Init. of AM-DIG Registers failed.\n",
		      board);
	  return error;
	}
      break;
    case ICP_AMPV:
      memcpy (&icp.dev[board][camera]->pv_cam,
	      &pv_def_cam, sizeof(struct pv_cam_t));
      if (camera == 0)
	{
	  if ((error = icp_ampv_init (board, 0)))
	    {
	      icp_printf ("icpci[%d]: Init. of PCVision Registers failed.\n",
			  board);
		return (error);
	    }
	}
      break;
    default:
      return -ENXIO;
    }
  }
  
  return OK;
}

/*
 * icp_mem_clear
 */

void
icp_mem_clear (int board, int mode)
{
  void *addr = NULL;
  char value;
  size_t size;
  
  ICP_PRINTF("icpci[%c]: icp_mem_clear.\n", 'F');
  
  switch (mode & ~ICP_ALL_BUFFERS)
    {
    case ICP_CLEAR_BLACK:
      value = 0x00;
      break;
    case ICP_CLEAR_GREY:
      value = 0x80;
      break;
    case ICP_CLEAR_WHITE:
      value = 0xFF;
      break;
    default:
      return;
    }
  
  if (mode & ICP_ALL_BUFFERS)
    {
      addr = icp.mem[board]->v_buf_base; 
      size = icp.mem[board]->buf_size;
    }
  else
    {
      addr = (icp.mem[board]->v_buf_base + 
	      icp.mem[board]->buffer[icp.mem[board]->cur_buf]);
      size = icp.mem[board]->raw_size;
    }
  
  icp_memset_io (board, addr, value, size);
}

/*
 * icp_mem_copy
 */

int
icp_mem_copy (int board, int camera, void *buffer)
{
  void *src_addr, *dst_addr;
  off_t src_off;
  size_t size; 
  short	width, height;
  int line;
  
  ICP_PRINTF("icpci[%c]: icp_mem_copy.\n", 'F');
  
  width = icp.mem[board]->width << cr_get_pixsz (board);
  height = icp.mem[board]->height;
  size = icp.mem[board]->raw_size;

  if (icp.bma_on)
    {
      src_addr = icp.bma[board]->v_buf_base;
      src_off = icp.bma[board]->buffer[icp.bma[board]->cur_buf];
    }
  else
    {
      src_addr = icp.mem[board]->v_buf_base;
      src_off = icp.mem[board]->buffer[icp.mem[board]->cur_buf];
    }

  if ((dst_addr = icp_malloc (size, PAGED)) == NULL)
    return -ENOMEM;
  
  if (icp.bma_on)
    memcpy (dst_addr, src_addr + src_off, size);    
  else
    {
      if (cr_get_smstat (board) == CR_ILACE)
	switch (cr_get_fstrt (board))
	  {
	  case CR_FIELD_EVEN:
	    for (line = 0; line < height / 2; line++)
	      {
		icp_memcpy_fromio (dst_addr + line * 2 * width,
				   src_addr + src_off + line * width, width);
		icp_memcpy_fromio (dst_addr + (line * 2 + 1) * width,
				   src_addr + src_off + line *
				   width + size / 2, width);
	      }
	    break;
	  case CR_FIELD_ODD:
	    for (line = 0; line < height / 2; line++)
	      {
		icp_memcpy_fromio (dst_addr + line * 2 * width,
				   src_addr + src_off + line *
				   width + size / 2, width);
		icp_memcpy_fromio (dst_addr + (line * 2 + 1) * width,
				   src_addr + src_off + line * width, width);
	      }
	    break;
	  }
      else
	icp_memcpy_fromio (dst_addr, src_addr + src_off, size);
    }

  icp_copyout (buffer, dst_addr, size);
  icp_free (dst_addr, PAGED);
  
  return OK;
}

/*
 * icp_mem_init
 */

int
icp_mem_init (int board) {
  
  int buffer;
  
  ICP_PRINTF("icpci[%c]: icp_mem_init.\n", 'F');
  
  /* first time initialization */
  if (icp.mem[board] == NULL)
    {
      /* allocate structure */
      icp.mem[board] =
	(struct icp_buf_t*)icp_malloc (sizeof (struct icp_buf_t), UNPAGED);
      if (icp.mem[board] == NULL)
	return -ENOMEM;
      memset (icp.mem[board], 0x00, sizeof(struct icp_buf_t));
      
      /* get buffer begin and size */
      icp.mem[board]->p_buf_base = icp.pci[board]->address3;
      icp.mem[board]->v_buf_base = icp.map[board]->v_mem_base;
      if (icp.map[board]->big_mem) {
	icp.mem[board]->buf_size = MEM_MAP_SIZE << 1;
      } else {
	icp.mem[board]->buf_size = MEM_MAP_SIZE;
      }

      icp_mem_clear (board, ICP_CLEAR_BLACK | ICP_ALL_BUFFERS);
    }
  
  /* switch to module specific subroutines for width / height / interlaced */
  switch (icp.cap[board]->module)
    {
    case ICP_AMVS:
      icp.mem[board]->width = vs_get_hsz (board);
      icp.mem[board]->height = vs_get_vsz (board);
      if (vs_get_smode (board) == VS_ILACE)
	icp.mem[board]->height <<= 1;
      break;
    case ICP_AMDG:
      icp.mem[board]->width = dg_get_hact (board);
      if (dg_get_ilpixel (board) == DG_ILEAVE)
      icp.mem[board]->width <<= 1;
      icp.mem[board]->height = dg_get_vact (board);
      if (dg_get_ilmode (board) == DG_ILACE)
	icp.mem[board]->height <<= 1;
      break;
    case ICP_AMPV:
      icp.mem[board]->width =
	pv_get_hactive (board) - pv_get_hoffset (board) + 2;
      icp.mem[board]->height =
	pv_get_vactive (board) - pv_get_voffset (board);
      if (pv_get_smode (board) == PV_ILACE)
	icp.mem[board]->height <<= 1;
      break;
    default:
      return -ENXIO;
    }
  
  /* calculate image size */ 
  icp.mem[board]->raw_size = icp.mem[board]->width * icp.mem[board]->height;
  if (icp.pci[board]->revision > ICP_REV_A &&
      cr_get_pixsz(board) == CR_PIXSZ_24_24)
    icp.mem[board]->raw_size *= 3;	/* 24bit storage at new boards */
  else
    icp.mem[board]->raw_size <<= cr_get_pixsz (board);
  
  ICP_PRINTF("icpci[%d]: width: %d, height: %d, pixel: %d.\n",
	     board, icp.mem[board]->width, icp.mem[board]->height,
	     icp.mem[board]->raw_size / icp.mem[board]->width / 
	     icp.mem[board]->height);
  ICP_PRINTF("icpci[%d]: size: %d.\n", board, icp.mem[board]->raw_size); 
  
  /* validate size; calculate number of available buffers */
  if (icp.mem[board]->raw_size > icp.mem[board]->buf_size)
    {
      icp_printf("icpci[%c]: At least one mem buffer have to be available.\n",
		 'E');
      return -ENOMEM;
    }

  /* align buffers to 4k (hardware limit) */
  icp.mem[board]->paged_size =
    (((icp.mem[board]->raw_size) + (1 << CR_AQSTART_SHIFT) - 1) &
     ~((1 << CR_AQSTART_SHIFT) - 1));
  icp.mem[board]->buffers =
    icp.mem[board]->buf_size / icp.mem[board]->paged_size;
  if (icp.mem[board]->buffers > ICP_BUFFERS_MAX)
    {
      icp_printf("icpci[%c]: Too much buffers (increase ICP_BUFFERS_MAX).\n",
		 'W');
      icp.mem[board]->buffers = ICP_BUFFERS_MAX;
    }
  
  /* assign start address of buffers */
  for (buffer = 0; buffer < icp.mem[board]->buffers; buffer++)
    icp.mem[board]->buffer[buffer] = buffer * icp.mem[board]->paged_size;
  ICP_PRINTF("icpci[%d]: No. of mem buffers: %d.\n", board,
	     icp.mem[board]->buffers); 
  ICP_PRINTF("icpci[%d]:               with: [%d pages].\n",
	     board, icp.mem[board]->paged_size / (1 << CR_AQSTART_SHIFT)); 
  
  return OK;
}  

/*
 * icp_bma_clear
 */

void
icp_bma_clear (int board, int mode)
{
  void *addr;
  char value;
  size_t size;
  
  ICP_PRINTF("icpci[%c]: icp_bma_clear.\n", 'F');
  
  if (!icp.bma_on)
    return;

  switch (mode & ~ICP_ALL_BUFFERS)
    {
    case ICP_CLEAR_BLACK:
      value = 0x00;
      break;
    case ICP_CLEAR_GREY:
      value = 0x80;
      break;
    case ICP_CLEAR_WHITE:
      value = 0xFF;
      break;
    default:
      return;
    }
  
  if (mode & ICP_ALL_BUFFERS)
    {
      addr = icp.bma[board]->v_buf_base; 
      size = icp.bma[board]->buf_size;
    }
  else
    {
      addr = (icp.bma[board]->v_buf_base + 
	      icp.bma[board]->buffer[icp.bma[board]->cur_buf]);
      size = icp.bma[board]->raw_size;
    }
  
  memset (addr, value, size);
}

/*
 * icp_bma_init
 */

int
icp_bma_init (int board)
{

  int _board, buffer;

  ICP_PRINTF("icpci[%c]: icp_bma_init.\n", 'F');
  
  if (!icp.bma_on)
    return -ENXIO;
  
  /* first time initialization */
  if (icp.bma[board] == NULL)
    {
      
      if (board == 0)
	{
	  /* split available memory */
	  icp.pieces = 0;
	  for (_board = 0; _board < icp.boards; _board++)
	    if (icp.map[_board]->big_mem)
	      /* 4 Mb */
	      icp.pieces += 2, icp.weight[_board] = 2;
	    else
	      /* 2 Mb */
	      icp.pieces++, icp.weight[_board] = 1;
	}

      /* allocate structure if necessary */
      icp.bma[board] =
	(struct icp_buf_t*)icp_malloc (sizeof (struct icp_buf_t), UNPAGED);
      if (icp.bma[board] == NULL)
	return -ENOMEM;
      memset (icp.bma[board], 0x00, sizeof(struct icp_buf_t));
      
      /* fill in structure */
      icp.bma[board]->buf_size =
	ROUND_DOWNTO_PAGE(icp.bma_size / icp.pieces * icp.weight[board]);
      icp.bma[board]->p_buf_base = icp.p_bma_base;
      icp.bma[board]->v_buf_base = icp.v_bma_base;

      for (_board = 0; _board < board; _board++)
	{
	  icp.bma[board]->p_buf_base += icp.bma[_board]->buf_size;
	  icp.bma[board]->v_buf_base += icp.bma[_board]->buf_size;
	}
      icp_bma_clear (board, ICP_CLEAR_BLACK | ICP_ALL_BUFFERS);
    }
  
  /* get image dimensions */
  icp.bma[board]->width = icp.mem[board]->width;
  if (icp.pci[board]->revision > ICP_REV_A)
    icp.bma[board]->width >>= cr_get_hdec (board);
  icp.bma[board]->height = icp.mem[board]->height;
  if (icp.pci[board]->revision > ICP_REV_A)
    icp.bma[board]->height >>= cr_get_vdec (board);
  icp.bma[board]->raw_size = icp.bma[board]->width * icp.bma[board]->height;
  if (icp.pci[board]->revision > ICP_REV_A &&
      cr_get_pixsz(board) == CR_PIXSZ_24_24)
    icp.bma[board]->raw_size *= 3;	/* 24bit storage at new boards */
  else
    icp.bma[board]->raw_size <<= cr_get_pixsz (board);


  if (icp.bma[board]->raw_size > icp.bma[board]->buf_size)
    {
      icp_printf("icpci[%c]: At least one bma buffer have to be available.\n",
		 'E');
      return -ENOMEM;
    }
  icp.bma[board]->raw_size += sizeof(struct icp_info_t);
  if (icp.bma[board]->raw_size > icp.bma[board]->buf_size)
    {
      icp_printf("icpci[%c]: Not enough buffer space to allocate "
		 "info structure.\n", 'E');
      return -ENOMEM;
    }
  
  /* adapt to page size */
  icp.bma[board]->paged_size = ROUND_UPTO_PAGE(icp.bma[board]->raw_size);
  
  /* clear validate size; calculate number of available buffers */
  icp.bma[board]->buffers =
    icp.bma[board]->buf_size / icp.bma[board]->paged_size;

  if (icp.bma[board]->buffers < icp.mem[board]->buffers)
    {
      icp_printf("icpci[%c]: You need at least %d bma buffers.\n", 'E',
		 icp.bma[board]->buffers);
      return -ENOMEM;
    }
  if (icp.bma[board]->buffers > ICP_BUFFERS_MAX)
    {
      icp_printf("icpci[%c]: Too much buffers (increase ICP_BUFFERS_MAX).\n",
		 'W');
      icp.bma[board]->buffers = ICP_BUFFERS_MAX;
    }
  
  /* assign start address of buffers */
  for (buffer = 0; buffer < icp.bma[board]->buffers; buffer++)
    icp.bma[board]->buffer[buffer] = buffer * icp.bma[board]->paged_size;
  
  ICP_PRINTF("icpci[%d]: No. of bma buffers: %d.\n",
	     board, icp.bma[board]->buffers);
  ICP_PRINTF("icpci[%d]:               with: [%ld pages].\n",
	     board, icp.bma[board]->paged_size / PAGE_SIZE); 
  
  /* setup registers for bma */
  /* disable busmaster requests */
  if (cr_set_bmstren (board, CR_DISABLE) < 0)
    {
      icp_error (board, "cr_set_bmstren");
      return -EIO;
    }
  if (icr_set_bmreq (board, ICR_DISABLE) < 0)
    {
      icp_error (board, "icr_set_bmreq");
      return -EIO;
    }
  /* set source values */ 
  if (cr_get_smstat (board) == CR_ILACE)
    {
      /* interlaced */
      if (cr_set_bmilace (board, CR_ENABLE) < 0)
	{
	  icp_error (board, "cr_set_bmilace");
	  return -EIO;
	}
    }
  else
    {
      /* not interlaced */
      if (cr_set_bmilace (board, CR_DISABLE) < 0)
	{
	  icp_error (board, "cr_set_bmilace");
	  return -EIO;
	}
    }
  
  if (cr_set_pitch (board, (icp.bma[board]->width << cr_get_pixsz(board)))
      < 0)
    {
      icp_error (board, "cr_set_pitch");
      return -EIO;
    }
  if (cr_set_aoix (board, (icp.bma[board]->width << cr_get_pixsz(board)) - 8)
      < 0)
    {
      icp_error (board, "cr_set_aoix");
      return -EIO;
    }
  
  return OK;
}

/*
 * icp_bma_copy
 */

int
icp_bma_copy (int board)
{

  u_long dst_addr;
  off_t	src_off, dst_off;
  size_t size; 
  short	width, height;

#if(0)
  ICP_PRINTF("icpci[%c]: icp_bma_copy.\n", 'F');
#endif

  width = icp.mem[board]->width << cr_get_pixsz (board);
  height = icp.mem[board]->height;
  size = icp.mem[board]->raw_size;

  if (cr_get_fcnt (board) > 1)
    /* if we use multiple snap, the buffers are CONTIGNUOUS */
    src_off = icp.mem[board]->raw_size * icp.mem[board]->cur_buf;
  else
    /* if we use singe snap, the buffers is ALIGNED */
    src_off = icp.mem[board]->buffer[icp.mem[board]->cur_buf];
  
  /* FIXME: for linux we use reserved upper memory */
  dst_addr = icp_virt_to_bus ((void*)icp.bma[board]->p_buf_base);
  dst_off = icp.bma[board]->buffer[icp.bma[board]->cur_buf];
  
  cr_set_bmstren (board, CR_DISABLE);
  icr_set_bmreq (board, ICR_DISABLE);      
  
  if (cr_get_smstat (board) == CR_ILACE)
    switch (cr_get_fstrt (board))
      {
      case CR_FIELD_EVEN:
	cr_set_bmsf (board, src_off);
	cr_set_bmss (board, src_off + size / 2);
	break;
      case CR_FIELD_ODD:
	cr_set_bmsf (board, src_off + size / 2);
	cr_set_bmss (board, src_off);
	break;
      }
  else
    cr_set_bmsf (board, src_off);
  
  icr_clear_fifoclr (board);
  cr_set_bmstren (board, CR_ENABLE);
  icr_set_bmdst (board, dst_addr + dst_off);
  icr_set_bmxc (board, size);
  icr_set_bmreq (board, ICR_ENABLE);

  return OK;
}

