/*		 ICP-PCI Kernel module free public release
 *
 *		       Copyright (C) 1996-99 GOM mbH.
 *	         Rebenring 33, 38106 Braunschweig, GERMANY
 *		            All rights reserved
 *
 * $Id: icpciMain.c,v 5.1.2.6 1999/08/27 15:39:45 matthias Exp $
 *
 */

/*
 * Defines:
 */

#ifdef DEBUG_ICPCI
#  define DEBUG_ICPCI_IRQ
#endif
#ifdef DEBUG_ICPCI_IRQ
#  define ICP_IRQ_PRINTF(fmt) icp_irq_printf (fmt)
#else
#  define ICP_IRQ_PRINTF(fmt)
#endif

/*
 * Includes:
 */

#include "icpciOS.h"	/* os specific stuff */
#include "icpciExt.h"	/* offical driver interface */
#include "icpciInt.h"	/* driver internals */
#include "icpciReg.h"	/* register definitions IC-PCI */
#include "amvsReg.h"	/* register definitions AM-VS */
#include "amdigReg.h"	/* register definitions AM-DIG */
#include "ampcvReg.h"	/* register definitions PCVision */

/*
 * Globals:
 */

/* main driver structure (adjust to BOARDS_MAX/DEVICES_MAX) */
struct icp_global_t icp =
{
  0,	/* boards */
  { NULL, NULL, NULL, NULL },	/* pci */
  { NULL, NULL, NULL, NULL },	/* map */
  { NULL, NULL, NULL, NULL },	/* cap */
  { NULL, NULL, NULL, NULL },	/* mem */
  {
    { NULL, NULL, NULL, NULL },	/* dev[0] */
    { NULL, NULL, NULL, NULL },	/* dev[1] */
    { NULL, NULL, NULL, NULL },	/* dev[2] */
    { NULL, NULL, NULL, NULL },	/* dev[3] */
  },
  { NULL, NULL, NULL, NULL },	/* bma */
  {
    { 0UL, 0UL, 0UL, 0UL, 0UL, 0UL },	/* acc[0] */
    { 0UL, 0UL, 0UL, 0UL, 0UL, 0UL },	/* acc[1] */
    { 0UL, 0UL, 0UL, 0UL, 0UL, 0UL },	/* acc[2] */
    { 0UL, 0UL, 0UL, 0UL, 0UL, 0UL }	/* acc[3] */
  },
  { NULL, NULL, NULL, NULL },	/* osp */
  /* device driver state */
  { ICP_INIT_TMO, ICP_INIT_TMO, ICP_INIT_TMO, ICP_INIT_TMO },	/* expire */
  { FALSE, FALSE, FALSE, FALSE },	/* timeout */
  { ICP_OP_NO, ICP_OP_NO, ICP_OP_NO, ICP_OP_NO },	/* opmode */
  { TRUE, TRUE, TRUE, TRUE },	/* complete */
  { 0, 0, 0, 0 },	/* count */
  { TRUE, TRUE, TRUE, TRUE },	/* first_read */
  { 0, 0, 0, 0 },	/* old_frames */
  { FALSE, FALSE, FALSE, FALSE },	/* opmode_err */
  /* io-port section */
  { FALSE, FALSE, FALSE, FALSE },	/* read_lock */
  { FALSE, FALSE, FALSE, FALSE },	/* write_lock */
  { NULL, NULL, NULL, NULL },	/* inp_buffer */
  { 0, 0, 0, 0 },	/* inp_count */
  { 0, 0, 0, 0 },	/* inp_index */
  /* global bma stuff */
  FALSE,/* bma_on */
  0UL,	/* p_bma_base */
  0UL,	/* bma_size */
  NULL,	/* v_bma_base */
  0,	/* pieces */
  { 1, 1, 1, 1 },	/* weight */
  { OK, OK, OK, OK }	/* errno */
};

/*
 * Prototypes:
 */

/*
 * icp_intr
 */

#define GRAB_FRAME_HANDLED() \
	{ \
	  ICP_IRQ_PRINTF("^q^"); \
	  if (frame_info != NULL) \
	    icp_fill_info (board, (void*)frame_info); \
	  icp_wakeup (board, CAMERA); \
	  icp_kill_proc (board); \
	}

#define SNAP_FRAME_HANDLED() \
	if (icp.count[board] >= cr_get_fcnt (board)) \
	  { \
	    icp.count[board] = 0; \
	    icp.complete[board] = TRUE; \
	    cr_set_fcnt (board, 1); \
	    cr_set_aqstart (board, icp.mem[board]-> \
			    buffer[icp.mem[board]->frst_buf]); \
	    icp.mem[board]->cur_buf = icp.mem[board]->frst_buf; \
	    if (icp.bma_on) \
	      icp.bma[board]->cur_buf = icp.bma[board]->frst_buf; \
	    GRAB_FRAME_HANDLED(); \
	  }

void
icp_intr (int irq)
{

  /* this runs with ALL disabled irq */

  int board;

  u_long irqs[ICP_BOARDS_MAX] = { 0UL, 0UL, 0UL, 0UL };
  
  struct icp_info_t *frame_info = NULL;

  board = 0;
  while (board < icp.boards)
    {
      /* test and clear all irq's on all boards at one time */
      if (icp.pci[board]->irq_line == irq)
	{
	  xcr_test_irqs (board, &irqs[board]);
	  if (irqs[board])
	    xcr_clear_irqs (board);
	}
      board++;
    }
  
  board = 0;
  while (board < icp.boards)
    {
      /* for all boards */
      if (irqs[board])
	{
#ifdef DEBUG_ICPCI_IRQ
	  icp_printf (" i%d", board);
#endif 
	  if (icp.bma_on)
	    frame_info =
	      (struct icp_info_t*)(icp.bma[board]->v_buf_base +
				   icp.bma[board]->
				   buffer[icp.bma[board]->cur_buf] +
				   icp.mem[board]->raw_size);

	  /* bus master done irq */
	  if (irqs[board] & ICR_BINT)
	    {
	      ICP_IRQ_PRINTF("b");
	      switch (icp.opmode[board])
		{
		case ICP_OP_GRAB:
		  ICP_IRQ_PRINTF("G");
		  GRAB_FRAME_HANDLED();
		  break;
		case ICP_OP_SNAP: 
		  ICP_IRQ_PRINTF("S");
		  SNAP_FRAME_HANDLED();
		  break;
		}
	    }
	
	  if (irqs[board] & ICR_MINT)
	    {
	      ICP_IRQ_PRINTF(">");
	      if (irqs[board] & CR_VBINT)
		ICP_IRQ_PRINTF("v");
	      if (irqs[board] & CR_OVBINT)
		ICP_IRQ_PRINTF("o");
	      if (irqs[board] & CR_EVBINT)
		ICP_IRQ_PRINTF("e");
	      /* verical blank irq's */
	      if (irqs[board] & CR_VBINT ||
		  irqs[board] & CR_OVBINT || irqs[board] & CR_EVBINT)
		switch (icp.opmode[board])
		  {
		  case ICP_OP_GRAB:
		    if (irqs[board] & CR_OVBINT)
		      if (cr_get_fstrt (board) == CR_FIELD_EVEN)
			icp_irq_printf ("\nicpci[%c]: odd irq - "
					"even starting field?\n", 'E');
		    if (irqs[board] & CR_EVBINT)
		      if (cr_get_fstrt (board) == CR_FIELD_ODD)
			icp_irq_printf ("\nicpci[%c]: even irq - "
					"odd starting field?\n", 'E');
		    ICP_IRQ_PRINTF("G");

		    icp.acc[board].frames_captured++;
		    if (icp_task_assigned (board))
		      {
			ICP_IRQ_PRINTF("|t|");
			if (!icp_check_blocked (board))
			  {
			    icp.acc[board].frames_valid++;
			    if (icp.bma_on)
			      {
				ICP_IRQ_PRINTF("[b]");
				icp_bma_copy (board);
			      }
			    else
			      GRAB_FRAME_HANDLED();
			  }
			else
			  icp.acc[board].frames_lost++;
		      }
		    break;
		    
		  case ICP_OP_SNAP:
		    if (irqs[board] & CR_OVBINT)
		      if (cr_get_fstrt (board) == CR_FIELD_EVEN)
			icp_irq_printf ("\nicpci[%c]: odd irq - "
					"even starting field?\n", 'E');
		    if (irqs[board] & CR_EVBINT)
		      if (cr_get_fstrt (board) == CR_FIELD_ODD)
			icp_irq_printf ("\nicpci[%c]: even irq - "
					"odd starting field?\n", 'E');
		    ICP_IRQ_PRINTF("S");

		    if (!icp.complete[board])
		      if (cr_get_vbstat (board) == CR_OUT_VB)
			icp_irq_printf ("\nicpci[%c]: test newaq out of VB?\n",
					'W');
		    if (cr_get_newaq (board) == CR_DISABLE)
		      {
			icp.count[board]++;
			if (cr_get_tmode (board) == CR_DISABLE)
			  {
			    ICP_IRQ_PRINTF("N");
			    if (icp.count[board] >= cr_get_fcnt (board))
			      break;
			  }
			else
			  {
			    ICP_IRQ_PRINTF("T");
			    if (cr_get_smstat (board) == CR_ILACE)
			      {
				if (icp.count[board] == 1 ||
				    icp.count[board] > 2)
				  break;
			      }
			    else
			      if (icp.count[board] > 1)
				break;
			  }

			if (icp.bma_on)
			  {
			    ICP_IRQ_PRINTF("[b]");
			    icp_bma_copy (board);
			  }
			else
			  SNAP_FRAME_HANDLED();

			if (cr_get_fcnt (board) > 1 && 
			    icp.count[board] < cr_get_fcnt (board))
			  {
			    if (icp.mem[board]->cur_buf <=
				icp.mem[board]->buffers)
			      {
				ICP_IRQ_PRINTF("+");
				icp.mem[board]->cur_buf++;
			      }
			    if (icp.bma[board]->cur_buf <=
				icp.bma[board]->buffers)
			      {
				ICP_IRQ_PRINTF("+");
				icp.bma[board]->cur_buf++;
			      }
			  }
		      }
		    break;

		  case ICP_OP_VB_GRAB:
		    ICP_IRQ_PRINTF("VG");
		    if (cr_get_gstat (board) == CR_ENABLE)
		      {
			ICP_IRQ_PRINTF("^q^");
			icp_wakeup (board, CAMERA);
		      }
#ifdef DEBUG_ICPCI
		    if (cr_get_vbstat (board) == CR_IN_VB)
		      icp_irq_printf ("\nicpci[%c]: test grab status in VB?\n",
				      'W');
#endif
		    break;
		  
		  case ICP_OP_VB_SNAP:
		    ICP_IRQ_PRINTF("VS");
		    ICP_IRQ_PRINTF("^q^");
		    icp_wakeup (board, CAMERA);
		    break;
		  
		  case ICP_OP_VB_FREEZ:
		    ICP_IRQ_PRINTF("VF");
		    ICP_IRQ_PRINTF("^q^");
		    icp_wakeup (board, CAMERA);
		    break;
		  }
	  
	      /* acquisition complete irq */
	      if (irqs[board] & CR_AQINT)
		{
		  ICP_IRQ_PRINTF("a");
		  switch (icp.opmode[board])
		    {
		    case ICP_OP_SNAP:
		      ICP_IRQ_PRINTF("S");
		      if (icp.bma_on)
			{
			  ICP_IRQ_PRINTF("[b]");
			  icp_bma_copy (board);
			}
		      else
			SNAP_FRAME_HANDLED();
		      break;
		    }
		}
	    
	      /* trigger irq */
	      if (irqs[board] & CR_TRIGINT)
		{
		  ICP_IRQ_PRINTF("t");
		  switch (icp.opmode[board])
		    {
		    case ICP_OP_TRIGGER:
		      ICP_IRQ_PRINTF("T");
		      icp_splhigh (board);
		      if (cr_get_vbstat (board) == CR_OUT_VB)
			{
			  if (cr_get_smstat (board) == CR_ILACE)
			    {
			      if (cr_get_fldstat (board) == CR_IN_ODD_FIELD)
				{
				  cr_set_fldsel (board, CR_ODD_FIELD);
				  cr_set_evbien (board, CR_DISABLE);
				  cr_set_ovbien (board, CR_ENABLE);
				}
			      else
				{
				  cr_set_fldsel (board, CR_EVEN_FIELD);
				  cr_set_evbien (board, CR_ENABLE);
				  cr_set_ovbien (board, CR_DISABLE);
				}
			    }
#ifdef DEBUG_ICPCI
			  else
			    {
			      icp_irq_printf ("\nicpci[%c]: sample current "
					      "field in VB?\n", 'W');
			    }
#endif
			}
		      icp_spllow (board);
		      cr_set_tmode (board,CR_DISABLE);
		      cr_set_acqmd (board, CR_SNAP);
		      icp.opmode[board] = ICP_OP_SNAP;
		    }
		}
	      
	      /* Input Port Interrupt */
	      if (irqs[board] & CR_INPINT)
	        {
		  ICP_IRQ_PRINTF("p");
		  icp.acc[board].events_captured++;
		  if (icp.inp_buffer[board] != NULL &&
		      icp.inp_count[board] != 0)
		    {
		      ICP_IRQ_PRINTF("C");
		      icp.acc[board].bytes_valid++;
#ifdef DEBUG_ICPCI_IRQ
		      icp_irq_printf (" %d ", icp.inp_index[board]);
#endif
		    if (icp.inp_index[board] < icp.inp_count[board])
		      icp.inp_buffer[board][icp.inp_index[board]++] =
			cr_get_inport (board);
		    else
		      {
			ICP_IRQ_PRINTF("^q^");
			icp_wakeup (board, IOPORT);
		      }
		    }
		  else
		    icp.acc[board].bytes_lost++;
		}
	    
	      /* PCVision interrupts */
	      if (irqs[board] & PV_SOTINT)
		ICP_IRQ_PRINTF ("\nicpci[I]: PCVision strt of trigger irq!\n");
	      if (irqs[board] & PV_EOTINT)
		ICP_IRQ_PRINTF ("\nicpci[I]: PCVision end of trig irq!\n");
	      if (irqs[board] & PV_EIN0INT)
		ICP_IRQ_PRINTF ("\nicpci[I]: PCVision external irq 0!\n");
	      if (irqs[board] & PV_EIN1INT)
		ICP_IRQ_PRINTF ("\nicpci[I]: PCVision external irq 1!\n");
	    }
	
	  /* master/target abort */
	  if (irqs[board] & ICR_MAINT)
	    icp_irq_printf ("\nicpci[%c]: master abort!\n", 'E');
	  if (irqs[board] & ICR_TAINT)
	    icp_irq_printf ("\nicpci[%c]: target abort!\n", 'E');
	  ICP_IRQ_PRINTF("\n");
	}
      board++;
    }
}

/*
 * icp_timeout
 */

void
icp_timeout (int board)
{

  /* this runs with ALL disabled irq's */

  ICP_PRINTF("icpci[%c]: icp_timeout.\n", 'F');

  icp.timeout[board] = TRUE;

  switch (icp.opmode[board])
    {
    case ICP_OP_SNAP:
      cr_set_acqmd (board, CR_FREEZE);
      icp.count[board] = 0;
      icp.complete[board] = TRUE;
    case ICP_OP_TRIGGER:
      icp.mem[board]->last_buf =
	icp.mem[board]->frst_buf + cr_get_fcnt (board);
      if (icp.bma_on)
	{
	  icp.bma[board]->last_buf =
	    icp.bma[board]->frst_buf + cr_get_fcnt (board);
	}

      cr_set_fcnt (board, 1);
      cr_set_aqstart (board, icp.mem[board]->
		      buffer[icp.mem[board]->frst_buf]);
      icp.mem[board]->cur_buf = icp.mem[board]->frst_buf;
      if (icp.bma_on)
	icp.bma[board]->cur_buf = icp.bma[board]->frst_buf;
      /* clear buffers not in critical section! */
      icp_marktq (board);
      break;

    case ICP_OP_GRAB:
    case ICP_OP_VB_SNAP:
    case ICP_OP_VB_GRAB:
      if (cr_set_acqmd (board, CR_FREEZE) < 0)
	icp_error (board, "cr_set_freeze");
      icp.mem[board]->frst_buf = icp.mem[board]->cur_buf;
      icp.mem[board]->last_buf = icp.mem[board]->frst_buf + 1;
      if (icp.bma_on)
	{
	  icp.bma[board]->frst_buf = icp.bma[board]->cur_buf;
	  icp.bma[board]->last_buf = icp.bma[board]->frst_buf + 1;
	}
      /* clear buffers not in critical section! */
      icp_marktq (board);
      break;

    case ICP_OP_VB_FREEZ:
    case ICP_OP_OTHER:
      icp_wakeup (board, CAMERA);
      break;
      
    default:
      icp_wakeup (board, CAMERA);
    }
}

/*
 * icp_task_queue
 */

void
icp_task_queue (int board)
{

  int buffer;

  /* this runs at scheduling time */

  ICP_PRINTF("icpci[%c]: icp_task_queue.\n", 'F');
  
  if (icp.bma_on)
    {
      for (buffer = icp.bma[board]->frst_buf;
	   buffer < icp.bma[board]->last_buf; buffer++)
	  {
	    icp.bma[board]->cur_buf = buffer;
	    icp_bma_clear (board, ICP_CLEAR_GREY);
	  }
      icp.bma[board]->cur_buf = icp.bma[board]->frst_buf;
    }

  for (buffer = icp.mem[board]->frst_buf;
       buffer < icp.mem[board]->last_buf; buffer++)
    if (buffer < icp.mem[board]->buffers)
      {
	icp.mem[board]->cur_buf = buffer;
	icp_mem_clear (board, ICP_CLEAR_GREY);
      }
  icp.mem[board]->cur_buf = icp.mem[board]->frst_buf;

  icp_wakeup (board, CAMERA);
}

/*
 * icp_open
 */

int
icp_open (struct icp_devno_t *devno, icp_flags_t flags)
{

  int error;

  int switch_camera = 0;

  ICP_PRINTF("icpci[%c]: icp_open.\n", 'F');
  
  /* the io-port stuff is handled in a seperate file */
  if (devno->module == ICP_AMNO)
    {
      if (icp.pci[devno->board]->revision > ICP_REV_A)
        return icp_iop_open (devno->board, flags);
      else
	return -ENXIO;
    }

  /* test for lock */
  if (devno->camera == ICP_CAM_IOCTL)
    {
      if (flags & O_EXCL)
	/* in ioctl-mode test only,if O_EXLC is set */
	if (icp.cap[devno->board]->cam_open > 0)
	  return -EBUSY;
    }
  else
    if (icp.cap[devno->board]->cam_open > 0)
      return -EBUSY;

  /* first open -> initialize */
    if (icp.cap[devno->board]->cam_open == 0)
      {
	/* camera section */
	/* camera set by minor device number? - else fall back to cam 0 */
	if (devno->camera != ICP_CAM_IOCTL)
	  switch_camera = devno->camera;

	/* init board with assigned camera structure and activate port */
	switch (icp.cap[devno->board]->module)
	  {
	  case ICP_AMVS:
	    if (icp_amvs_init (devno->board, switch_camera))
	      return -EIO;
	    if (vs_set_camsel (devno->board, switch_camera))
	      {
		icp_error (devno->board, "vs_set_camsel");
		return -EIO;
	      }
	    break;
	  case ICP_AMDG:
	    if (icp_amdg_init (devno->board, switch_camera) < 0)
	      return -EIO;
	    /* no switching, only cam 0 available */
	    break;
	  case ICP_AMPV:
	    if (icp_ampv_init (devno->board, switch_camera) < 0)
	      return -EIO;
	    if (pv_set_camsel (devno->board, switch_camera) < 0)
	      {
		icp_error (devno->board, "pv_set_camsel");
		return -EIO;
	      }
	  break;
	  default:
	    return -ENXIO;
	  }
	icp.cap[devno->board]->cur_cam = switch_camera;	/* update structure */
	
	/* mem buffers section */
	/* initialize mem buffers */
	if ((error = icp_mem_init (devno->board)))
	  return error;

	/* mem-buffer 0 should always be available */
	if (cr_set_aqstart (devno->board, icp.mem[devno->board]->buffer[0]))
	  {
	    icp_error (devno->board, "cr_set_aqstart");
	    return -EIO;
	  }
	icp.mem[devno->board]->cur_buf = 0;	/* update structure */
	icp_mem_clear (devno->board, ICP_CLEAR_BLACK);
	
	/* bma buffers section (if available) */
	if (icp.bma_on)
	  {
	    /* initialize bma buffers */
	    if ((error = icp_bma_init (devno->board)))
	      return error;
	    /* buffer 0 should always be available */      
	    icp.bma[devno->board]->cur_buf = 0;	/* update structure */
	    icp_bma_clear (devno->board, ICP_CLEAR_BLACK);
	  }

	/* initialize signal and pid (no pid assigned, no signal delivered) */
	icp_deassign_signal (devno->board);

	/* clear task assignment */
	icp_deassign_task (devno->board);

	/* set default operation mode */
	icp.opmode[devno->board] = ICP_OP_NO;

	/* enable main interrupts */
	icp_enable_irq (devno->board, ICR_MINT | ICR_BINT);
      }

    /* lock device now */
    icp.cap[devno->board]->cam_open++;
    
    return OK;
}

/*
 * icp_close
 */

int
icp_close (struct icp_devno_t *devno, icp_flags_t flags)
{

  /* the io-port stuff is handled in a seperate file */
  if (devno->module == ICP_AMNO)
    {
      if (icp.pci[devno->board]->revision > ICP_REV_A)
        return icp_iop_close (devno->board, flags);
      else
	return -ENXIO;
    }

  /* if last process, close the device */
  if (icp.cap[devno->board]->cam_open == 1)
    {
      /* disable grabbing */
      if (cr_get_acqmd (devno->board) != CR_FREEZE)
	cr_set_acqmd (devno->board, CR_FREEZE);
      
      /* cleanup signal type and target */
      icp_deassign_signal (devno->board);
      /* clear task assignment */
      icp_deassign_task (devno->board);
      
      /* disable busmaster requests */
      if (icp.bma_on)
	{
	  cr_set_bmstren (devno->board, CR_DISABLE);
	  icr_set_bmreq (devno->board, ICR_DISABLE);
	}
      
      /* disable main interrupts */
      icp_disable_irq (devno->board, CR_VBINT | CR_OVBINT | CR_EVBINT |
		       CR_AQINT | CR_TRIGINT |
		       ICR_MINT | ICR_BINT);
      
      /* force sane state */
      icp_del_timer (devno->board, CAMERA);
      icp.first_read[devno->board] = TRUE;
      icp.complete[devno->board] = TRUE;
      
      /* reset operation mode */
      icp.opmode[devno->board] = ICP_OP_NO;
    }
  
  /* unlock devices now */
  if (icp.cap[devno->board]->cam_open > 0)
    icp.cap[devno->board]->cam_open--;

  return OK;
}

/*
 * icp_read
 */

icp_ssize_t
icp_read (struct icp_devno_t *devno, icp_flags_t flags,
	  char *buffer, icp_size_t count)
{
  int error;
  
  ICP_PRINTF("icpci[%c]: icp_read %d byte(s).\n", 'F', count);

  /* the io-port stuff is handled in a seperate file */
  if (devno->module == ICP_AMNO)
    {
      if (icp.pci[devno->board]->revision > ICP_REV_A)
        return icp_iop_read (devno->board, flags, buffer, count);
      else
	return -ENXIO;
    }

  /* argument validation */
  if (buffer == NULL || count != icp.mem[devno->board]->raw_size)
    return -EINVAL;

  /* initilaize 'complete' status, if needed */
  if (icp.first_read[devno->board])
    {
      icp.first_read[devno->board] = FALSE;
      icp.complete[devno->board] = FALSE;
    }

  /* save current buffer status */
  icp.mem[devno->board]->frst_buf = icp.mem[devno->board]->cur_buf;
  if (icp.bma_on)
    icp.bma[devno->board]->frst_buf = icp.bma[devno->board]->cur_buf;

  /* if not complete */
  if (!icp.complete[devno->board])
    {
      icp.timeout[devno->board] = FALSE;
      /* enable interrupt */
      if ((error = icp_enable_irq (devno->board, CR_VB_FLD | CR_AQINT)))
	return error;
      /* set frame_count (only one frame) */
      if (cr_set_fcnt (devno->board, 1))
	{
	  icp_error (devno->board, "cr_set_fcnt");
	  return -EIO;
	}
    /* enable snapping */
      if (cr_set_acqmd (devno->board, CR_SNAP))
	{
	  icp_error (devno->board, "cr_set_acq");
	  return -EIO;
	}
      icp.opmode[devno->board] = ICP_OP_SNAP;
    }

  /* set timeout and wait or continue */
  if (flags & O_NONBLOCK)
    /* set timer BUT don't wait on queue */
    icp_add_timer (devno->board, icp.expire[devno->board] * 2, CAMERA);
  else
    /* set timer AND wait on queue */
    if ((error = icp_sleep (devno->board,
			    icp.expire[devno->board] * 2, CAMERA)))
      return error; /* -ERESTARTSYS */
  
  /* if complete */ 
  if (icp.complete[devno->board])
    {
      icp_mem_copy (devno->board, devno->camera, buffer);
      icp.first_read[devno->board] = TRUE;
      /* copy to user space */
      icp.opmode[devno->board] = ICP_OP_NO;
      if (icp.timeout[devno->board])
	{
	  icp.timeout[devno->board] = FALSE;
	  return -ENODATA;
	}
    }
  
  return icp.mem[devno->board]->raw_size;
}

/*
 * icp_write
 */

icp_ssize_t
icp_write (struct icp_devno_t *devno, icp_flags_t flags,
           const char *buffer, icp_size_t count)
{

  ICP_PRINTF("icpci[%c]: icp_write %d byte(s).\n", 'F', count);

  /* write is only implemented for the io-ports at newer boards */
  if (devno->module != ICP_AMNO)
    return -ENXIO;
  if (icp.pci[devno->board]->revision < ICP_REV_B)
    return -ENXIO;

  return icp_iop_write (devno->board, buffer, count);
}  

/*
 * icp_mmap
 */

int
icp_mmap (struct icp_devno_t *devno, u_long *base, size_t size)
{

  ICP_PRINTF("icpci[%c]: icp_mmap.\n", 'F');

  /* can't never mmap the io-ports */
  if (devno->module == ICP_AMNO)
    return -ENXIO;

  /* since the memory - buffers cannot be page aligned (multi-snap),
     mmap is allowed only on the ioctl device */
  if (devno->camera != ICP_CAM_IOCTL)
      return -ENXIO;
  
  /* argument validation */
  if (size % PAGE_SIZE)
    return -EFAULT;
  
  if (devno->module == ICP_MOD_IDENT)
    {
      if (size > icp.bma[devno->board]->buf_size)
	return -EFBIG;
      /* the module special/camera special device
	 mmap's the upper memory space */
      *base = icp.bma[devno->board]->p_buf_base;
      ICP_PRINTF("icpci[%c]: mapping on-board memory %#lx.\n", 'F', *base);
    }
  else
    {
      if (size > icp.mem[devno->board]->buf_size)
	return -EFBIG;
      /* the normal module/camera special device
	 mmap's the on-board memory space */
      *base = icp.mem[devno->board]->p_buf_base;
      ICP_PRINTF("icpci[%c]: mapping upper memory %#lx.\n", 'F', *base);
    }

  return OK;
}
  
/*
 * icp_select
 */

int
icp_select (struct icp_devno_t *devno, int type)
{
  int result;
  
  int new_frames;
  
  void *src_addr, *dst_addr;
  off_t	src_off, dst_off;
  
  result = 0;
  new_frames = 0;

  /* can't yet select on io-ports */
  if (devno->module == ICP_AMNO)
    return -ENXIO;

  if (type & ICP_SEL_IN)
    switch (icp.opmode[devno->board])
      {
      case ICP_OP_GRAB:
	icp.opmode_err[devno->board] = FALSE;
	new_frames = icp.acc[devno->board].frames_valid;
	if (new_frames > icp.old_frames[devno->board])
	  {
	    icp.old_frames[devno->board] = new_frames;
	    result |= ICP_SEL_IN;
	    break;
	  }
	icp.old_frames[devno->board] = new_frames;
	icp_assign_task (devno->board);
	break;
	
      case ICP_OP_SNAP:
      case ICP_OP_TRIGGER:
	icp.opmode_err[devno->board] = FALSE;
	if (icp.complete[devno->board])
	  {
	    icp.opmode[devno->board] = ICP_OP_NO;
	    result |= ICP_SEL_IN;
	    break;
	  }
	break;
	
      default:
	/* return immediately, but report the error in exception section */ 
	icp.opmode_err[devno->board] = TRUE;
      }
  
  /* exeption means ENODATA */ 
  if (type & ICP_SEL_EX)
    {
      if (icp.opmode_err[devno->board])
	{
	  icp.opmode_err[devno->board] = FALSE;
	  icp_printf ("icpci[%c]: opmode error!\n", 'E');
	  result |= ICP_SEL_EX;
	}
      else
	if (icp.complete[devno->board])
	  {
	    if (icp.timeout[devno->board])
	      {
		icp.timeout[devno->board] = FALSE;
		icp_printf ("icpci[%c]: timeout snap cycle!\n", 'E');
		return ICP_SEL_EX;
	      }
	    else
	      {
		/* here we test only the current buffer */
		src_addr = icp.mem[devno->board]->v_buf_base;
		src_off = (icp.mem[devno->board]->
			   buffer[icp.mem[devno->board]->cur_buf]);
		if (icp_readl (((u_long*)(src_addr + src_off))) ==
		    ICP_MEM_TEST) {
		  icp_printf ("icpci[%c]: mem buffer %d not valid!\n", 'E',
			      icp.mem[devno->board]->cur_buf);
		}
		dst_addr = icp.bma[devno->board]->v_buf_base;
		dst_off = (icp.bma[devno->board]->
			   buffer[icp.bma[devno->board]->cur_buf]);
		if (*((u_long*)(dst_addr + dst_off)) == ICP_BMA_TEST) {
		  icp_printf ("icpci[%c]: bma buffer %d not valid!\n", 'E',
			      icp.bma[devno->board]->cur_buf);
		}
	      }
	  }
    }
  
  return result;
}


